import logging
from typing import Iterator, Callable
from buff163_unofficial_api.rest_adapter import RestAdapter
from buff163_unofficial_api.exceptions import Buff163Exception
from buff163_unofficial_api.models import *


class Buff163API:
    def __init__(
        self,
        hostname: str = "buff.163.com/api",
        session_cookie: str = "",
        ssl_verify: bool = True,
        logger: logging.Logger = None,
        page_size: int = 20,
    ):
        """Buff163API default constructor.

        Args:
            hostname (str, optional): API url. Defaults to "buff.163.com/api".
            session_cookie (str, optional): Personal session cookie (like an api token). Defaults to "".
            ssl_verify (bool, optional): Set to false if having SSL/TLS cert validation issues. Defaults to True.
            logger (logging.Logger, optional): App logger. Defaults to None.
            page_size (int, optional): Items per page. Defaults to 20.
        """
        self._rest_adapter = RestAdapter(hostname, session_cookie, ssl_verify, logger)
        self._page_size = page_size

    def get_featured_market_item(self) -> Item:
        """Get first featured market item (random).

        Returns:
            Item: Overview of specific item.
        """
        return self.get_featured_market()[0]

    def get_featured_market(self, pageNum: int = 1) -> List[Item]:
        """Get entire featured item.

        Args:
            pageNum (int, optional): Which page number to get. Defaults to 1.

        Returns:
            List[Item]: List of overview of items.
        """
        result = self._rest_adapter.get(
            endpoint=f"/market/goods?game=csgo&page_num={pageNum}"
        )
        market = [Item(**item) for item in result.data["data"]["items"]]
        return market

    def fetch_image_data(self, item: Item):
        """Fetches Item icon.

        Args:
            item (Item): Specific Item from market.
        """
        item.data = self._rest_adapter.fetch_data(url=item.goods_info.icon_url)

    def _page(
        self, endpoint: str, model: Callable[..., Model], max_amt: int = 80
    ) -> Iterator[Model]:
        """Pages through set number of pages.

        Args:
            endpoint (str): API endpoint requested.
            model (Callable[..., Model]): Specific model that will be paged.
            max_amt (int, optional): Max items to get from pages. Defaults to 80.

        Yields:
            Iterator[Model]: List of specific model.
        """
        amt_yielded = 0
        curr_page = last_page = 1
        ep_params = {
            "game": "csgo",
            "page_size": self._page_size,
        }

        # Keep fetching pages until the last page
        while curr_page <= last_page:
            ep_params["page_num"] = curr_page
            result = self._rest_adapter.get(endpoint=endpoint, ep_params=ep_params)
            data = result.data["data"]

            # Increment curr_page by 1 and update the last_page based on header info returned
            last_page = data["total_page"]
            curr_page = data["page_num"] + 1
            for datam in data["items"]:
                yield model(**datam)
                amt_yielded += 1
                if amt_yielded >= max_amt:
                    last_page = 0
                    break

    def get_featured_market_paged(self, max_amt: int = 80) -> Iterator[Item]:
        """Page the featured market

        Args:
            max_amt (int, optional): Amount of Items to get. Defaults to 80.

        Returns:
            _type_: List of Items

        Yields:
            Iterator[Item]: List of Items
        """
        return self._page(endpoint="/market/goods", model=Item, max_amt=max_amt)



from unittest import TestCase
from unittest.mock import MagicMock
from buff163_unofficial_api.buff163_unofficial_api import Buff163API
from buff163_unofficial_api.models import Item, Result


class TestBuff163API(TestCase):
    def setUp(self) -> None:
        self.buff163api = Buff163API(page_size=10)
        self.buff163api._rest_adapter = MagicMock()

    def test_get_featured_market_item_returns_one_item(self):
        self.buff163api._rest_adapter.get.return_value = Result(
            200, headers={}, data=[{"id": 1, "url": "example.com"}]
        )
        item = self.buff163api.get_featured_market_item()
        self.assertIsInstance(item, Item)

    def test_get_featured_market_returns_list_of_items(self):
        item_amt = 2
        self.buff163api._rest_adapter.get.return_value = Result(
            200,
            headers={},
            data=[
                {"id": 1, "url": "example.com"},
                {"id": 2, "url": "example.com"},
                {"id": 3, "url": "example.com"},
            ],
        )
        item_list = self.buff163api.get_featured_market(amt=item_amt)
        self.assertIsInstance(item_list, list)
        self.assertTrue(len(item_list), item_amt)
        self.assertIsInstance(item_list[0], Item)

    def test_get_featured_market_paged_returns_iterator_of_item(self):
        self.buff163api._rest_adapter.get.side_effect = [
            Result(
                200,
                headers={},
                data=[
                    {"id": 1, "url": "example.com"},
                    {"id": 2, "url": "example.com"},
                    {"id": 3, "url": "example.com"},
                ],
            ),
            Result(200, headers={}, data=[]),
        ]
        item_iterator = self.buff163api.get_featured_market_paged()
        item1 = next(item_iterator)
        item2 = next(item_iterator)
        item3 = next(item_iterator)
        self.assertIsInstance(item3, Item)
        with self.assertRaises(StopIteration):
            item4 = next(item_iterator)



import os
from typing import List, Dict, Union, TypeVar
from buff163_unofficial_api.exceptions import Buff163Exception

Model = TypeVar("Model", covariant=True)


class Result:
    def __init__(self, status_code: int, message: str = "", data: List[Dict] = None):
        """Result returned from low-level RestAdapter

        Args:
            status_code (int): The HTTP status code resulting from the API call.
            message (str, optional): Message providing additional info about result. Defaults to "".
            data (List[Dict], optional): Data returned from API call. Defaults to None.
        """
        self.status_code = int(status_code)
        self.message = str(message)
        self.data = data if data else []


class Exterior:
    def __init__(
        self, category: str, id: int, internal_name: str, localized_name: str
    ) -> None:
        self.category = category
        self.id = id
        self.internal_name = internal_name
        self.localized_name = localized_name


class Tags:
    def __init__(
        self,
        exterior: Union[Exterior, dict],
        quality: Union[Exterior, dict],
        rarity: Union[Exterior, dict],
        type: Union[Exterior, dict],
        weapon: Union[Exterior, dict],
    ) -> None:
        self.exterior = Exterior(**exterior) if isinstance(exterior, dict) else exterior
        self.quality = Exterior(**quality) if isinstance(quality, dict) else quality
        self.rarity = Exterior(**rarity) if isinstance(rarity, dict) else rarity
        self.type = Exterior(**type) if isinstance(type, dict) else type
        self.weapon = Exterior(**weapon) if isinstance(weapon, dict) else weapon


class Info:
    def __init__(self, tags: Tags) -> None:
        self.tags = tags


class GoodsInfo:
    def __init__(
        self,
        icon_url: str,
        info: Union[Info, dict],
        original_icon_url: str,
        steam_price: str,
        steam_price_cny: str,
        **kwargs,
    ) -> None:
        self.icon_url = icon_url
        self.info = Info(**info) if isinstance(info, dict) else info
        self.original_icon_url = original_icon_url
        self.steam_price = steam_price
        self.steam_price_cny = steam_price_cny
        self.__dict__.update(kwargs)


class Item:
    def __init__(
        self,
        buy_max_price: str,
        buy_num: int,
        can_bargain: bool,
        goods_info: Union[GoodsInfo, dict],
        id: int,
        market_hash_name: str,
        market_min_price: int,
        name: str,
        quick_price: str,
        sell_min_price: str,
        sell_num: int,
        sell_reference_price: str,
        short_name: str,
        steam_market_url: str,
        transacted_num: int,
        data: bytes = bytes(),
        **kwargs,
    ) -> None:
        self.buy_max_price = buy_max_price
        self.buy_num = buy_num
        self.can_bargain = can_bargain
        self.goods_info = (
            GoodsInfo(**goods_info) if isinstance(goods_info, dict) else goods_info
        )
        self.id = id
        self.market_hash_name = market_hash_name
        self.market_min_price = market_min_price
        self.name = name
        self.quick_price = quick_price
        self.sell_min_price = sell_min_price
        self.sell_num = sell_num
        self.sell_reference_price = sell_reference_price
        self.short_name = short_name
        self.steam_market_url = steam_market_url
        self.transacted_num = transacted_num
        self.data = data
        self.__dict__.update(kwargs)

    def save_icon_to(self, path: str = "./", file_name: str = ""):
        if not self.data:
            raise Buff163Exception("No data to save")
        try:
            save_file_name = file_name if file_name else f"{self.market_hash_name}.jpg"
            save_path = os.path.join(path, save_file_name)
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            with open(save_path, "wb") as f:
                f.write(self.data)

        except Exception as e:
            raise Buff163Exception(str(e)) from e



import requests
import requests.packages
from typing import Dict
import logging
from json.decoder import JSONDecodeError
from buff163_unofficial_api.exceptions import Buff163Exception
from buff163_unofficial_api.models import Result


class RestAdapter:
    def __init__(
        self,
        hostname: str = "buff.163.com/api",
        session_cookie: str = "",
        ssl_verify: bool = True,
        logger: logging.Logger = None,
    ) -> None:
        """Constructor for RestAdapter

        Args:
            hostname (str): Api url. Defaults to "buff.163.com/api".
            session_cookie (str, optional): Used for authentication. Defaults to "".
            ssl_verify (bool, optional): Set to false if having SSL/TLS cert validation issues. Defaults to True.
            logger (logging.Logger, optional): App logger. Defaults to None.
        """
        self._logger = logger or logging.getLogger(__name__)
        self.url = f"https://{hostname}"
        self._session_cookie = session_cookie
        self._ssl_verify = ssl_verify
        if not ssl_verify:
            # noinspection PyUnresolvedReferences
            requests.packages.urllib3.disable_warnings()

    def _do(
        self, http_method: str, endpoint: str, ep_params: Dict = None, data: Dict = None
    ) -> Result:
        """Private method for api requests (GET, POST, DELETE, etc.)

        Args:
            http_method (str): GET, POST, DELETE, etc.
            endpoint (str): URL endpoint
            ep_params (Dict, optional): Endpoint parameters. Defaults to None.
            data (Dict, optional): Data to pass to Buff163API. Defaults to None.

        Raises:
            Buff163Exception: Requests fail
            Buff163Exception: Bad JSON
            Buff163Exception: Error response code

        Returns:
            Result: status_code, message, data
        """
        full_url = self.url + endpoint
        headers = {"Cookie": self._session_cookie}

        log_line_pre = f"method={http_method}, url={full_url}, params={ep_params}"
        log_line_post = ", ".join(
            (log_line_pre, "success={}, status_code={}, message={}")
        )
        # Performing an HTTP request and logging its details; exceptions are logged and a custom exception is raised.
        try:
            self._logger.debug(msg=log_line_pre)
            response = requests.request(
                method=http_method,
                url=full_url,
                verify=self._ssl_verify,
                headers=headers,
                params=ep_params,
                json=data,
            )

        except requests.exceptions.RequestException as e:
            self._logger.error(msg=(str(e)))
            raise Buff163Exception("Request failed") from e

        # Convert JSON response to a Python object; raise and log a custom exception for JSON parsing errors
        try:
            data_out = response.json()
        except (ValueError, JSONDecodeError) as e:
            self._logger.error(msg=log_line_post.format(False, None, e))
            raise Buff163Exception("Bad JSON in response") from e

        # Check code is valid and data is produced
        is_login_error = data_out["code"] != "OK"
        if is_login_error:
            log_line = log_line_post.format(False, 401, "Login Required")
            self._logger.error(msg=log_line)
            raise Buff163Exception("Login is required")

        # Check response status code for success (200-299) and log accordingly.
        is_success = 299 >= response.status_code >= 200
        # log_line = log_line_post.format(
        #     is_success, response.status_code, response.reason
        # )
        if is_success:
            self._logger.debug(msg="log_line")
            return Result(response.status_code, message=response.reason, data=data_out)
        self._logger.error(msg=log_line)
        raise Buff163Exception(f"{response.status_code}: {response.reason}")

    def get(self, endpoint: str, ep_params: Dict = None) -> Result:
        """Sends a GET request to a specified API endpoint.

        Args:
            endpoint (str): The endpoint for the GET request.
            ep_params (Dict, optional): Parameters to include in request. Defaults to None.

        Returns:
            Result: status_code, message, data
        """
        return self._do(http_method="GET", endpoint=endpoint, ep_params=ep_params)

    def post(self, endpoint: str, ep_params: Dict = None, data: Dict = None) -> Result:
        """Sends a POST request to a specified API endpoint.

        Args:
            endpoint (str): The endpoint for the POST request.
            ep_params (Dict, optional): Parameters to include in the request. Defaults to None.
            data (Dict, optional): Data passed in the request. Defaults to None.

        Returns:
            Result: status_code, message, data
        """
        return self._do(
            http_method="POST", endpoint=endpoint, ep_params=ep_params, data=data
        )

    def delete(
        self, endpoint: str, ep_params: Dict = None, data: Dict = None
    ) -> Result:
        """Sends a DELETE request to a specified API endpoint.

        Args:
            endpoint (str): The endpoint for the POST request.
            ep_params (Dict, optional): Parameters to include in the request. Defaults to None.
            data (Dict, optional): Data passed in the request. Defaults to None.

        Returns:
            Result: status_code, message, data
        """
        return self._do(
            http_method="DELETE", endpoint=endpoint, ep_params=ep_params, data=data
        )

    def fetch_data(self, url: str) -> bytes:
        """Private method for fetching data from url.

        Args:
            url (str): Url of fetch request.

        Raises:
            Buff163Exception: Request failure.
            Buff163Exception: Status code not valid.

        Returns:
            bytes: Data in bytes (mainly for images).
        """
        http_method = "GET"
        try:
            log_line = f"method={http_method}, url={url}"
            self._logger.debug(msg=log_line)
            response = requests.request(
                method=http_method, url=url, verify=self._ssl_verify
            )
        except requests.exceptions.RequestException as e:
            self._logger.error(msg=(str(e)))
            raise Buff163Exception(str(e)) from e

        # If status_code in 200-299 range, return byte stream, otherwise raise exception
        is_success = 299 >= response.status_code >= 200
        log_line = f"success={is_success}, status_code={response.status_code}, message={response.reason}"
        self._logger.debug(msg=log_line)
        if not is_success:
            raise Buff163Exception(response.reason)
        return response.content



import requests
from requests.exceptions import RequestException
from unittest import TestCase, mock

from buff163_unofficial_api.exceptions import Buff163Exception
from buff163_unofficial_api.models import Result
from buff163_unofficial_api.rest_adapter import RestAdapter


class TestRestAdapter(TestCase):
    def setUp(self):
        self.rest_adapter = RestAdapter()
        self.response = requests.Response()

    # _do
    def test__do_good_request_returns_result(self):
        # Arrange
        self.response.status_code = 200
        self.response._content = "{}".encode()
        # Act
        with mock.patch("requests.request", return_value=self.response):
            result = self.rest_adapter._do("GET", "")
            # Assert
            self.assertIsInstance(result, Result)

    def test__do_bad_request_raises_buff163_exception(self):
        with mock.patch("requests.request", side_effect=RequestException):
            with self.assertRaises(Buff163Exception):
                self.rest_adapter._do("GET", "")

    def test__do_bad_json_raises_buff163_exception(self):
        bad_json = '{"invalid JSON": '
        self.response._content = bad_json
        with mock.patch("requests.request", return_value=self.response):
            with self.assertRaises(Buff163Exception):
                self.rest_adapter._do("GET", "")

    def test__do_300_or_higher_raises_buff163_exception(self):
        self.response.status_code = 300
        with mock.patch("requests.request", return_value=self.response):
            with self.assertRaises(Buff163Exception):
                self.rest_adapter._do("GET", "")

    def test__do_199_or_lower_raises_buff163_exception(self):
        self.response.status_code = 199
        with mock.patch("requests.request", return_value=self.response):
            with self.assertRaises(Buff163Exception):
                self.rest_adapter._do("GET", "")

    # get
    def test_get_method_passes_in_get(self):
        self.response.status_code = 200
        self.response._content = "{}".encode()
        with mock.patch("requests.request", return_value=self.response) as request:
            self.rest_adapter.get("")
            self.assertTrue(request.method, "GET")

    # post
    def test_post_method_passes_in_post(self):
        self.response.status_code = 200
        self.response._content = "{}".encode()
        with mock.patch("requests.request", return_value=self.response) as request:
            self.rest_adapter.post("")
            self.assertTrue(request.method, "POST")

    # delete
    def test_delete_method_passes_in_delete(self):
        self.response.status_code = 200
        self.response._content = "{}".encode()
        with mock.patch("requests.request", return_value=self.response) as request:
            self.rest_adapter.delete("")
            self.assertTrue(request.method, "DELETE")

    # def test_fetch_data(self):
    #     self.fail()



from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="buff163-unofficial-api",
    version="0.1.2",
    author="Mark Zhdan",
    author_email="markzhdan@gmail.com",
    description="An unofficial API wrapper for Buff163, a CS skin marketplace.",
    url="https://github.com/markzhdan/buff163-unofficial-api",
    license="MIT",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    long_description=long_description,
    long_description_content_type="text/markdown",
)
